@@ -31,6 +31,7 @@
 #include <sys/inotify.h>
 #include <glib.h>
 
+#include <connman/provision.h>
 #include "connman.h"
 
 struct connman_config_service {
@@ -48,6 +49,10 @@ struct connman_config_service {
 	char *private_key_passphrase_type;
 	char *phase2;
 	char *passphrase;
+	GSList *service_identifiers;
+	char *config_ident; /* file prefix */
+	char *config_entry; /* entry name */
+	connman_bool_t hidden;
 };
 
 struct connman_config {
@@ -65,6 +70,7 @@ static int inotify_wd = -1;
 
 static GIOChannel *inotify_channel = NULL;
 static uint inotify_watch = 0;
+static connman_bool_t cleanup = FALSE;
 
 #define INTERNAL_CONFIG_PREFIX           "__internal"
 
@@ -85,6 +91,7 @@ static uint inotify_watch = 0;
 #define SERVICE_KEY_IDENTITY           "Identity"
 #define SERVICE_KEY_PHASE2             "Phase2"
 #define SERVICE_KEY_PASSPHRASE         "Passphrase"
+#define SERVICE_KEY_HIDDEN             "Hidden"
 
 static const char *config_possible_keys[] = {
 	CONFIG_KEY_NAME,
@@ -106,6 +113,7 @@ static const char *service_possible_keys[] = {
 	SERVICE_KEY_IDENTITY,
 	SERVICE_KEY_PHASE2,
 	SERVICE_KEY_PASSPHRASE,
+	SERVICE_KEY_HIDDEN,
 	NULL,
 };
 
@@ -125,26 +133,53 @@ static void unregister_config(gpointer data)
 
 static void unregister_service(gpointer data)
 {
-	struct connman_config_service *service = data;
+	struct connman_config_service *config_service = data;
+	struct connman_service *service;
+	char *service_id;
+	GSList *list;
 
-	connman_info("Removing service configuration %s", service->ident);
+	if (cleanup == TRUE)
+		goto free_only;
 
-	protected_services = g_slist_remove(protected_services, service);
+	connman_info("Removing service configuration %s",
+						config_service->ident);
 
-	g_free(service->ident);
-	g_free(service->type);
-	g_free(service->name);
-	g_free(service->ssid);
-	g_free(service->eap);
-	g_free(service->identity);
-	g_free(service->ca_cert_file);
-	g_free(service->client_cert_file);
-	g_free(service->private_key_file);
-	g_free(service->private_key_passphrase);
-	g_free(service->private_key_passphrase_type);
-	g_free(service->phase2);
-	g_free(service->passphrase);
-	g_free(service);
+	protected_services = g_slist_remove(protected_services,
+						config_service);
+
+	for (list = config_service->service_identifiers; list != NULL;
+							list = list->next) {
+		service_id = list->data;
+
+		service = __connman_service_lookup_from_ident(service_id);
+		if (service != NULL) {
+			__connman_service_set_immutable(service, FALSE);
+			__connman_service_remove(service);
+		}
+
+		if (__connman_storage_remove_service(service_id) == FALSE)
+			DBG("Could not remove all files for service %s",
+								service_id);
+	}
+
+free_only:
+	g_free(config_service->ident);
+	g_free(config_service->type);
+	g_free(config_service->name);
+	g_free(config_service->ssid);
+	g_free(config_service->eap);
+	g_free(config_service->identity);
+	g_free(config_service->ca_cert_file);
+	g_free(config_service->client_cert_file);
+	g_free(config_service->private_key_file);
+	g_free(config_service->private_key_passphrase);
+	g_free(config_service->private_key_passphrase_type);
+	g_free(config_service->phase2);
+	g_free(config_service->passphrase);
+	g_slist_free_full(config_service->service_identifiers, g_free);
+	g_free(config_service->config_ident);
+	g_free(config_service->config_entry);
+	g_free(config_service);
 }
 
 static void check_keys(GKeyFile *keyfile, const char *group,
@@ -353,6 +388,12 @@ static int load_service(GKeyFile *keyfile, const char *group,
 		service->passphrase = str;
 	}
 
+	service->config_ident = g_strdup(config->ident);
+	service->config_entry = g_strdup_printf("service_%s", service->ident);
+
+	service->hidden = g_key_file_get_boolean(keyfile, group,
+						SERVICE_KEY_HIDDEN, NULL);
+
 	if (service_created)
 		g_hash_table_insert(config->service_table, service->ident,
 					service);
@@ -595,9 +636,20 @@ static gboolean inotify_data(GIOChannel *channel, GIOCondition cond,
 
 			config = g_hash_table_lookup(config_table, ident);
 			if (config != NULL) {
+				int ret;
+
 				g_hash_table_remove_all(config->service_table);
 				load_config(config);
-				__connman_service_provision_changed(ident);
+				ret = __connman_service_provision_changed(ident);
+				if (ret > 0) {
+					/*
+					 * Re-scan the config file for any
+					 * changes
+					 */
+					g_hash_table_remove_all(config->service_table);
+					load_config(config);
+					__connman_service_provision_changed(ident);
+				}
 			}
 		}
 
@@ -683,10 +735,14 @@ void __connman_config_cleanup(void)
 {
 	DBG("");
 
+	cleanup = TRUE;
+
 	remove_watch();
 
 	g_hash_table_destroy(config_table);
 	config_table = NULL;
+
+	cleanup = FALSE;
 }
 
 static char *config_pem_fsid(const char *pem_file)
@@ -714,7 +770,7 @@ static void provision_service(gpointer key, gpointer value, gpointer user_data)
 	struct connman_service *service = user_data;
 	struct connman_config_service *config = value;
 	struct connman_network *network;
-	const void *ssid;
+	const void *ssid, *service_id;
 	unsigned int ssid_len;
 
 	/* For now only WiFi service entries are supported */
@@ -739,9 +795,17 @@ static void provision_service(gpointer key, gpointer value, gpointer user_data)
 	if (memcmp(config->ssid, ssid, ssid_len) != 0)
 		return;
 
+	service_id = __connman_service_get_ident(service);
+	config->service_identifiers =
+		g_slist_prepend(config->service_identifiers,
+				g_strdup(service_id));
+
 	__connman_service_set_immutable(service, TRUE);
 
-	__connman_service_set_favorite(service, TRUE);
+	__connman_service_set_favorite_delayed(service, TRUE, TRUE);
+
+	__connman_service_set_config(service, config->config_ident,
+						config->config_entry);
 
 	if (config->eap != NULL)
 		__connman_service_set_string(service, "EAP", config->eap);
@@ -791,6 +855,13 @@ static void provision_service(gpointer key, gpointer value, gpointer user_data)
 
 	if (config->passphrase != NULL)
 		__connman_service_set_string(service, "Passphrase", config->passphrase);
+
+	if (config->hidden == TRUE)
+		__connman_service_set_hidden(service);
+
+	__connman_service_mark_dirty();
+
+	__connman_service_save(service);
 }
 
 int __connman_config_provision_service(struct connman_service *service)
@@ -819,10 +890,11 @@ int __connman_config_provision_service(struct connman_service *service)
 }
 
 int __connman_config_provision_service_ident(struct connman_service *service,
-							const char *ident)
+			const char *ident, const char *file, const char *entry)
 {
 	enum connman_service_type type;
 	struct connman_config *config;
+	int ret = 0;
 
 	DBG("service %p", service);
 
@@ -832,9 +904,132 @@ int __connman_config_provision_service_ident(struct connman_service *service,
 		return -ENOSYS;
 
 	config = g_hash_table_lookup(config_table, ident);
-	if(config != NULL)
+	if (config != NULL) {
+		GHashTableIter iter;
+		gpointer value, key;
+		gboolean found = FALSE;
+
+		g_hash_table_iter_init(&iter, config->service_table);
+
+		/*
+		 * Check if we need to remove individual service if it
+		 * is missing from config file.
+		 */
+		if (file != NULL && entry != NULL) {
+			while (g_hash_table_iter_next(&iter, &key,
+							&value) == TRUE) {
+				struct connman_config_service *config_service;
+
+				config_service = value;
+
+				if (g_strcmp0(config_service->config_ident,
+								file) != 0)
+					continue;
+
+				if (g_strcmp0(config_service->config_entry,
+								entry) != 0)
+					continue;
+
+				found = TRUE;
+				break;
+			}
+
+			DBG("found %d ident %s file %s entry %s", found, ident,
+								file, entry);
+
+			if (found == FALSE) {
+				/*
+				 * The entry+8 will skip "service_" prefix
+				 */
+				g_hash_table_remove(config->service_table,
+						entry + 8);
+				ret = 1;
+			}
+		}
+
 		g_hash_table_foreach(config->service_table,
 						provision_service, service);
+	}
 
-	return 0;
+	return ret;
+}
+
+struct connman_config_entry **connman_config_get_entries(void)
+{
+	GHashTableIter iter_file, iter_config;
+	gpointer value, key;
+	struct connman_config_entry **entries = NULL;
+	int i = 0, count;
+
+	g_hash_table_iter_init(&iter_file, config_table);
+	while (g_hash_table_iter_next(&iter_file, &key, &value) == TRUE) {
+		struct connman_config *config_file = value;
+
+		count = g_hash_table_size(config_file->service_table);
+
+		entries = g_try_realloc(entries, (i + count + 1) *
+					sizeof(struct connman_config_entry *));
+		if (entries == NULL)
+			return NULL;
+
+		g_hash_table_iter_init(&iter_config,
+						config_file->service_table);
+		while (g_hash_table_iter_next(&iter_config, &key,
+							&value) == TRUE) {
+			struct connman_config_service *config = value;
+
+			entries[i] = g_try_new0(struct connman_config_entry,
+						1);
+			if (entries[i] == NULL)
+				goto cleanup;
+
+			entries[i]->ident = g_strdup(config->ident);
+			entries[i]->name = g_strdup(config->name);
+			entries[i]->ssid = g_try_malloc0(config->ssid_len + 1);
+			if (entries[i]->ssid == NULL)
+				goto cleanup;
+
+			memcpy(entries[i]->ssid, config->ssid,
+							config->ssid_len);
+			entries[i]->ssid_len = config->ssid_len;
+			entries[i]->hidden = config->hidden;
+
+			i++;
+		}
+	}
+
+	if (entries != NULL) {
+		entries = g_try_realloc(entries, (i + 1) *
+					sizeof(struct connman_config_entry *));
+		if (entries == NULL)
+			return NULL;
+
+		entries[i] = NULL;
+
+		DBG("%d provisioned AP found", i);
+	}
+
+	return entries;
+
+cleanup:
+	connman_config_free_entries(entries);
+	return NULL;
+}
+
+void connman_config_free_entries(struct connman_config_entry **entries)
+{
+	int i;
+
+	if (entries == NULL)
+		return;
+
+	for (i = 0; entries[i]; i++) {
+		g_free(entries[i]->ident);
+		g_free(entries[i]->name);
+		g_free(entries[i]->ssid);
+		g_free(entries[i]);
+	}
+
+	g_free(entries);
+	return;
 }
